
# source: http://code.activestate.com/recipes/576694/
import collections


class OrderedSet(collections.MutableSet):

	def __init__(self, iterable=None):
		self.end = end = []
		end += [None, end, end]         # sentinel node for doubly linked list
		self.map = {}                   # key --> [key, prev, next]
		if iterable is not None:
			self |= iterable

	def __getstate__(self):
		return self.end, self.map

	def __setstate__(self, state):
		self.end, self.map = state

	def __hashkey__(self):
		return self.__class__.__name__, self.__getstate__()

	def __len__(self):
		return len(self.map)

	def __contains__(self, key):
		return key in self.map

	def add(self, key):
		if key not in self.map:
			end = self.end
			curr = end[1]
			curr[2] = end[1] = self.map[key] = [key, curr, end]

	def discard(self, key):
		if key in self.map:
			key, prev, next = self.map.pop(key)
			prev[2] = next
			next[1] = prev

	def __iter__(self):
		end = self.end
		curr = end[2]
		while curr is not end:
			yield curr[0]
			curr = curr[2]

	def __reversed__(self):
		end = self.end
		curr = end[1]
		while curr is not end:
			yield curr[0]
			curr = curr[1]

	def pop(self, last=True):
		if not self:
			raise KeyError('set is empty')
		key = self.end[1][0] if last else self.end[2][0]
		self.discard(key)
		return key

	def __repr__(self):
		if not self:
			return '%s()' % (self.__class__.__name__,)
		return '%s(%r)' % (self.__class__.__name__, list(self))

	def __eq__(self, other):
		if isinstance(other, OrderedSet):
			return len(self) == len(other) and list(self) == list(other)
		return set(self) == set(other)



from re import findall as re_findall
from json import loads as json_loads


def json_from_string(s):
    """
    :type s: str
    :rtype: dict
    """
    match = re_findall(r"{.+[:,].+}|\[.+[,:].+\]", s)
    return json_loads(match[0]) if match else None


import hashlib
import base64
import base32hex


def make_hashable(obj):
	try:
		return make_hashable(obj.__hashkey__())
	except AttributeError:
		if isinstance(obj, (tuple, list)):
			return tuple((make_hashable(e) for e in obj))

		if isinstance(obj, dict):
			return tuple(sorted((k, make_hashable(v)) for k,v in obj.items()))

		if isinstance(obj, (set, frozenset)):
			return tuple(sorted(make_hashable(e) for e in obj))

		return obj


def hash_dictionary(dictionary, simplify=False, base=64):
	"""
	:type dictionary: dict
	:type simplify: bool
	:type base: int
	:rtype: str
	"""
	if simplify:
		new_dict = {
			k: (v if isinstance(v, (int, float, str)) else str(v))
			for k, v in dictionary.items()
		}

	else:
		new_dict = {
			k: (v if isinstance(v, (int, float, str)) else hash_object(obj=v, base=base))
			for k, v in dictionary.items()
		}

	return str(set(new_dict.items()))


def hash_object(obj, base=64):
	"""
	:type obj: Obj
	:type base: int
	:rtype str
	"""
	if hasattr(obj, '__hash64__') and base == 64:
		return obj.__hash64__()
	elif hasattr(obj, '__hash32__') and base == 32:
		return obj.__hash32__()

	hash_maker = hashlib.sha256()
	hash_maker.update(repr(make_hashable(obj)).encode())
	if base == 64:
		return base64.b64encode(hash_maker.digest()).decode()
	elif base == 32:
		return base32hex.b32encode(hash_maker.digest()).replace('=', '-')
	else:
		raise ValueError(f'base{base} is unknown!')


def hash_file(path, base=64, block_size=65536):
	hash_maker = hashlib.sha256()
	with open(path, 'rb') as file:
		block = file.read(block_size)
		while len(block) > 0:
			hash_maker.update(block)
			block = file.read(block_size)
	if base == 64:
		return base64.b64encode(hash_maker.digest()).decode()
	elif base == 32:
		return base32hex.b32encode(hash_maker.digest()).replace('=', '-')
	else:
		raise ValueError(f'base{base} is unknown!')



from .Dictionary import Dictionary
from .has_duplicates import has_duplicates, get_duplicates
from .get_dict_product import get_dict_product
from .is_iterable import is_iterable
from .get_intersection import get_intersection
from .rename_dict_keys import rename_dict_keys
from .apply_on_dict_keys import apply_on_dict_keys
from .merge_dictionaries import merge_dictionaries
from .OrderedSet import OrderedSet
from .remove_list_duplicates import remove_list_duplicates
from .serialize_arguments import serialize_arguments
from .flatten import flatten
from .sum import sum
from .Series import Series
from .create_grid import create_grid



def is_iterable(obj, include_string=False):
	"""
	:type include_string: bool
	:rtype: bool
	"""
	try:
		iterator = iter(obj)
		return (type(obj) is not str) or include_string
	except TypeError:
		return False





from itertools import product


def get_dict_product(d):
	labels = d.keys()
	value_lists = [d[k] for k in d.keys()]
	value_combinations = product(*value_lists)
	result = [dict(zip(labels, values)) for values in value_combinations]
	return result





# source: https://github.com/slezica/python-frozendict/blob/master/frozendict/__init__.py
import collections

try:
	from collections import OrderedDict
except ImportError:  # python < 2.7
	OrderedDict = NotImplemented

iteritems = getattr(dict, 'iteritems', dict.items)  # py2-3 compatibility


class ImmutableDictionary(collections.Mapping):
	"""
	An immutable wrapper around dictionaries that implements the complete :py:class:`collections.Mapping`
	interface. It can be used as a drop-in replacement for dictionaries where immutability is desired.
	"""

	dict_cls = dict

	def __init__(self, *args, **kwargs):
		self._dict = self.dict_cls(*args, **kwargs)
		self._hash = None

	def __getitem__(self, key):
		return self._dict[key]

	def __contains__(self, key):
		return key in self._dict

	def copy(self, **add_or_replace):
		return self.__class__(self, **add_or_replace)

	def __iter__(self):
		return iter(self._dict)

	def __len__(self):
		return len(self._dict)

	def __repr__(self):
		return f'<{self.__class__.__name__}>\n{repr(self._dict)}'

	def __hash__(self):
		if self._hash is None:
			h = 0
			for key, value in iteritems(self._dict):
				h ^= hash((key, value))
			self._hash = h
		return self._hash


class ImmutableOrderedDict(ImmutableDictionary):
	"""
	A ImmutableDictionary subclass that maintains key order
	"""

	dict_cls = OrderedDict


if OrderedDict is NotImplemented:
	del ImmutableOrderedDict


from copy import deepcopy


def rename_dict_keys(dictionary, mapping, recursive=True, raise_error=False):
	"""
	:type dictionary: dict
	:type mapping: dict
	:type recursive: bool
	:type raise_error: bool
	:rtype: dict
	"""
	dictionary = deepcopy(dictionary)

	if isinstance(dictionary, dict):
		for old_key, new_key in mapping.items():
			if old_key in dictionary:
				dictionary[new_key] = dictionary.pop(old_key)
			elif raise_error:
				raise KeyError(f'{old_key} does not exist in dictionary.')
		if recursive:
			for key, value in dictionary.items():
				dictionary[key] = rename_dict_keys(dictionary=value, mapping=mapping)

		return dictionary

	elif isinstance(dictionary, list):
		if recursive:
			return [rename_dict_keys(dictionary=d, mapping=mapping) for d in dictionary]
		else:
			return dictionary

	else:
		return dictionary




from math import ceil, floor


def trim(string, max_length=None, cut_from='middle'):
	"""
	:param str string: string to be shortened
	:param int or NoneType max_length: if None, the string will only be stripped of leading and trailing white spaces
	:param str cut_from: example: 'what a wonderful world!'    end --> 'what ...', middle --> 'wh ... ld', start --> ' ... world'
	:rtype: str
	"""
	string = str(string).strip()
	if max_length is None:
		return string

	max_length = int(max_length)
	cut_from = cut_from.lower()
	if len(string) > max_length:
		if cut_from == 'end':
			return string[:max_length - 4] + ' ...'
		elif cut_from == 'start':
			return '... ' + string[len(string) - max_length + 4:]
		elif cut_from == 'middle':
			length_allowed = max_length - 5
			left_side = ceil(length_allowed / 2)
			right_side = floor(length_allowed / 2)
			return string[:left_side] + ' ... ' + string[len(string) - right_side:]
	else:
		return string


def get_intersection(l1, l2):
	""" Finds the intersection of 2 lists including common duplicates"""

	intersection_set = set(l1).intersection(set(l2))
	intersection_list = []
	for i in intersection_set:
		num = min(l1.count(i), l2.count(i))
		for j in range(num):
			intersection_list.append(i)
	return intersection_list





from urllib.request import urlopen
from .json_from_string import json_from_string


def json_from_url(url):
	"""
	:type url: str
	:rtype: dict
	"""
	with urlopen(url=url) as url:
		data = json_from_string(url.read().decode())
	return data


def sum(l):
	"""
	:type l: list
	"""
	if len(l) == 1:
		return l[0]
	elif len(l) == 2:
		return l[0] + l[1]
	elif len(l) > 2:
		middle = len(l) // 2
		return sum(l[:middle]) + sum(l[middle:])
	else:
		raise TypeError('list is empty!')



class Series:
	def __init__(self, start, end, step):
		self.start = start
		self.end = end
		self.step = step

	def generate(self):
		num = self.start
		while num <= self.end:
			yield num
			num += self.step

	def __getstate__(self):
		return (self.start, self.end, self.step)

	def __setstate__(self, state):
		self.start, self.end, self.step = state

	def __hashkey__(self):
		return (self.__class__.__name__, self.__getstate__())

	@property
	def list(self):
		return list(self.generate())



from .timestamp import TimeStamp
from .distance import get_distance
from .textfile_to_string import textfile_to_string
from .remove_non_alphanumeric import remove_non_alphanumeric
from .split_and_strip import split_and_strip
from .colour import colour
from .get_size import get_size
from .trim import trim
from .get_function_arguments import get_function_arguments



from .ImmutableDictionary import ImmutableOrderedDict, ImmutableDictionary
from ..hash import hash_object

from datetime import datetime
from inspect import isfunction
from collections import OrderedDict
from functools import wraps
from copy import deepcopy

IMMUTABLE_TYPES = frozenset({
	type(None),
	bool,
	int,
	float,
	complex,
	str,
	bytes,
	tuple,
	datetime,
	ImmutableDictionary,
	ImmutableOrderedDict
})

def is_immutable(obj):
	"""
	Determine whether a thing is immutable
	:param value obj: a Python value (object or primitive)
	:returns true if the value is known to be immutable
	:rtype bool
	"""

	return (
		any([
			isinstance(obj, t)
			for t in IMMUTABLE_TYPES.union(frozenset({Immutable, frozenset}))
		]) or
		isfunction(obj)
	)

class Immutable(object):
	"""
		A wrapper for arbitrary Python objects,
		that protects against attribute mutation
		Warning: `freezemethod` and `freezeobject` perform a `deepcopy`
		of the object's attributes upon each method call in order to
		test whether the method call would modify the object.
		The attribute copy may use a significant amount of memory.
		To save memory, model data with structures designed to be immutable,
		such as tuples, frozensets, ImmutableDictionaries, etc.
	"""

	def __init__(self, value):
		"""
			Construct a immutable object
			:param value value: a Python object
			:returns a immutable version of the object
		"""

		self._original_object = value

	def __hashkey__(self):
		return 'immutable', self.__class__.__name__, self._original_object

	@property
	def original_object(self):
		return deepcopy(self._original_object)

	def __getattribute__(self, name):
		"""
			Override attribute getters,
			returning immutable versions of each attribute
			:param str name: an attribute name
			:returns value: a wrapped version of the attribute
		"""
		if name in ['_original_object', 'original_object', '__getstate__', '__setstate__', '__call__', '__hashkey__']:
			return super(Immutable, self).__getattribute__(name)
		if name == '__class__':
			return getattr(self._original_object, name)

		attribute = getattr(self._original_object, name)

		# protect methods from mutation
		if callable(attribute) and not isfunction(attribute):
			#print(f'"{name}" is callable, making function immutable', end=' ... ')
			return make_function_immutable(self._original_object, name)
		# protect non-method attributes from mutation
		else:
			#print(f'"{name}" is not callable, making immutable', end=' ... ')
			return make_immutable(attribute)

	def __setattr__(self, name, value):
		"""
			Override attribute setters,
			preventing attribute mutation
			:param str name: an attribute name
			:param value value: a value for the attribute to take
			:raises LiquidNitrogenException in most cases
		"""

		if name == '_original_object':
			super(Immutable, self).__setattr__('_original_object', value)
		else:
			print(f'Cannot alter attribute {name} of object {self} of type {type(self)}')
			raise AttributeError(f'Cannot alter attribute {name} of object {self} of type {type(self)}')

	def __call__(self, *args, **kwargs):
		"""
			Forwards method calls to the wrapped object
			:param list args: any positional call arguments
			:param dict kwargs: any named call arguments
			:returns value: the return value of the wrapped method call
			:raises LiquidNitrogenException if call would modify this object
		"""
		return self._original_object.__call__(*args, **kwargs)

	def __repr__(self):
		"""
			Forwards string representation calls to the wrapped object
			:returns str: a string representation of the wrapped object
		"""

		return self._original_object.__repr__()

	def __getitem__(self, item):
		return self._original_object[item]

	def __dir__(self):
		return dir(self._original_object)

	def __getstate__(self):
		return self._original_object

	def __setstate__(self, state):
		self._original_object = state

def make_function_immutable(obj, name):
	"""
		Create an immutable version of a method,
		that detects when a call would modify the object,
		and instead raises LiquidNitrogenException
		Warning: `freezemethod` and `freezeobject` perform a `deepcopy`
		of the object's attributes upon each method call in order to
		test whether the method call would modify the object.
		The attribute copy may use a significant amount of memory.
		To save memory, model data with structures designed to be immutable,
		such as tuples, frozensets, ImmutableDictionaries, etc.
		:param object obj: an object
		:param str name: a method to wrap on the object
		:returns method: a wrapped method that protects against mutation
	"""

	obj_copy_hash = hash_object(obj)
	method_copy = getattr(obj, name)

	@wraps(method_copy)
	def protected_method(*args, **kwargs):
		result = method_copy(*args, **kwargs)

		# Has the copy mutated compared to the original?
		if obj_copy_hash == hash_object(obj):
			return result
		else:
			print(f'immutable method call {name} with arguments {args} {kwargs} would mutate {obj}')
			raise RuntimeError(f'immutable method call {name} with arguments {args} {kwargs} would mutate {obj}')

	return protected_method


def make_immutable(obj):
	"""
		Given a Python value, return an immutable version of the value.
		Natively immutable values such as tuples, immutablesets, and ImmutableDictionaries
		are simply returned as-is.
		Methods and objects are wrapped to raise LiquidNitrogenExceptions
		on attempts to set attributes or call methods
		that would mutate the objects.
		Warning: `freezemethod` and `freezeobject` perform a `deepcopy`
		of the object's attributes upon each method call in order to
		test whether the method call would modify the object.
		The attribute copy may use a significant amount of memory.
		To save memory, model data with structures designed to be immutable,
		such as tuples, immutablesets, ImmutableDictionaries, etc.
		:param value obj: a Python value (object or primitive)
		:returns a immutable version of the value
	"""

	if is_immutable(obj):
		return obj
	elif isinstance(obj, list):
		return tuple(obj)
	elif isinstance(obj, set):
		return frozenset(obj)
	elif isinstance(obj, OrderedDict):
		return ImmutableOrderedDict(obj)
	elif isinstance(obj, dict):
		return ImmutableDictionary(obj)
	else:
		return Immutable(obj)



import inspect

def serialize_arguments(exclude=None):
    """
    when put inside a function, it produces a dictionary of all arguments passed into the function
    :param list[str] or NoneType exclude: arguments to exclude
    :rtype: list[tuple]
    """
    exclude = exclude or []
    frame = inspect.currentframe().f_back
    args, _, _, values = inspect.getargvalues(frame)
    result = []
    for arg in args:
        if arg not in exclude:
            result.append((arg, values[arg]))
    return result


from collections import Mapping, Iterable
from sys import getsizeof

def deep_getsizeof(o, ids):
	"""Find the memory footprint of a Python object
	This is a recursive function that rills down a Python object graph
	like a dictionary holding nested ditionaries with lists of lists
	and tuples and sets.
	The sys.getsizeof function does a shallow size of only. It counts each
	object inside a container as pointer only regardless of how big it
	really is.
	:param o: the object
	:param ids:
	:return:
	"""
	ids = ids or set()
	d = deep_getsizeof
	if id(o) in ids:
		return 0

	r = getsizeof(o)
	ids.add(id(o))

	if isinstance(o, str):
		return r

	if isinstance(o, Mapping):
		return r + sum(d(k, ids) + d(v, ids) for k, v in o.items())

	if isinstance(o, Iterable):
		return r + sum(d(x, ids) for x in o)
	# source: https://github.com/the-gigi/deep/blob/master/deeper.py#L80
	return r

def get_size(obj, exclude_objects=None):
	""" Find the memory footprint of a Python object recursively
	:param object obj:
	:rtype: int
	"""
	exclude_objects = exclude_objects or []
	return deep_getsizeof(o=obj, ids=set([id(x) for x in exclude_objects]))